import traceback as tb
from colored import fg, bg, attr

def save_traceback():
    try:
        raise TracebackException("Could not register Variable.")
    except TracebackException as e:
        full_tb = tb.extract_stack()

        # ignore this very function in the traceback
        return full_tb[:-1]

class RegisterError(Exception):

    def __str__(self):
        base_exc = super().__str__()
        return base_exc + ("\n\n"+fg('red')+"The variable definition occured in"+attr('reset')+":\n"""+format_traceback_list(self.traceback) if self.traceback is not None else "")

    def __init__(self, msg='', traceback=None, *args, **kwargs):

        # storing the traceback which provides useful information about where the exception occurred
        self.traceback = traceback

        super().__init__(msg)

class TracebackException(Exception):
    pass

def format_traceback_list(traceback_list, ignore_miniflask=True, exc=None):
    if ignore_miniflask:
        traceback_list = [t for t in traceback_list if not t.filename.endswith("src/miniflask/event.py") and (not t.filename.endswith("src/miniflask/miniflask.py") or t.name in ["load","run"])]

    # ignore "raise" lines
    if exc is not None and "raise" in traceback_list[-1].line:
        t = traceback_list[-1]
        t.filename = fg('green')+t.filename+attr('reset')
        t.lineno = fg('yellow')+str(t.lineno)+attr('reset')
        t.name = fg('blue')+attr("bold")+t.name+attr('reset')
        exception_type = fg('red')+attr('bold')+type(exc).__name__+attr('reset')
        last_msg = "  File %s, line %s, in %s\n    %s: %s" % (t.filename, t.lineno, t.name, exception_type, str(exc))
        traceback_list = traceback_list[:-1]
    if exc is not None and not "raise" in traceback_list[-1].line:
        exception_type = fg('red')+attr('bold')+type(exc).__name__+attr('reset')
        last_msg = "    %s: %s" % (exception_type, str(exc))
    else:
        last_msg = ""
    for t in traceback_list:
        t.filename = fg('green')+t.filename+attr('reset')
        t.lineno = fg('yellow')+str(t.lineno)+attr('reset')
        t.name = fg('blue')+attr("bold")+t.name+attr('reset')
    return "".join(tb.format_list(traceback_list))+last_msg



from .miniflask import miniflask as init, print_info
from .event import outervar
from .modules import *
from .state import like

# meta
__version__       = "1.21"
__title__         = "miniflask"
__description__   = "Small research-oriented hook-based plugin engine."
__url__           = "https://github.com/da-h/miniflask"
__uri__           = __url__
__doc__           = __description__ + " <" + __uri__ + ">"
__documentation__ = 'https://da-h.github.io/miniflask'
__source__        = __url__
__tracker__       = __url__ + '/issues'
__author__        = "David Hartmann"
__license__       = "MIT License"
__copyright__     = "Copyright (c) 2020 " + __author__




import sys
import os
from itertools import zip_longest
from miniflask.miniflask import highlight_module, highlight_val, highlight_name, highlight_val_overwrite, like

html_module = lambda x: x
html_name= lambda x: x
html_val = lambda x: x
html_val_overwrite = lambda x: x

def listsettings(state, asciicodes=True):

    # colors
    color_module = highlight_module if asciicodes else html_module
    color_name = highlight_name if asciicodes else html_name
    color_val = highlight_val if asciicodes else html_val
    color_val_overwrite = highlight_val_overwrite if asciicodes else html_val_overwrite
    linesep = os.linesep if asciicodes else "<br>"

    last_k = []
    if len(state.all) == 0:
        return "No Settings available."
    maxklen = max(len(k) for k in state.all.keys())
    text = "Folder│"+color_name("module")+"│"+color_module("variable")+(" "*(maxklen-22))+" = "+color_val("value")+linesep
    text += "—"*(maxklen+8)+linesep
    for k, v in sorted(state.all.items()):

        # ignore state variables that are not registered for argument parsing
        if k not in state.default:
            continue

        klen = len(k)
        korig = k
        overwritten = v != (state.default[k].default if isinstance(state.default[k],like) else state.default[k])
        k = k.split(".")
        if len(k) > 1:
            k_hidden = [" "*len(ki) if ki==ki2 and asciicodes else ki for ki,ki2 in zip_longest(k,last_k) if ki is not None]
            last_k = k
            if k_hidden[-2] == "default":
                k_hidden[-3] = color_name(k_hidden[-3])
            else:
                k_hidden[-2] = color_name(k_hidden[-2])
            k_hidden[-1] = color_module(k_hidden[-1])
        else:
            k_hidden = k
            k_hidden[-1] = color_module(k_hidden[-1])

        append = "" if not overwritten else " ⟶   "+color_val_overwrite(str(v))
        text += "│".join(k_hidden)+(" "*(maxklen-klen))+" = "+color_val(str(state.default[korig]))+append+linesep

    return text


def init(state):
    print(listsettings(state))

def settings_html(state):
    html = "<code>"+listsettings(state, asciicodes=False)+"</code>"
    return html


def register(mf):
    mf.register_event("init", init)
    mf.register_event("settings_html", settings_html)



from miniflask.dummy import miniflask_dummy
from colored import fg, bg, attr
from miniflask.miniflask import highlight_module
import re

def color_module(moduleid, short=False):
    moduleid = moduleid.split(".")
    moduleid[0] = attr('dim')+moduleid[0]
    moduleid[-1] = attr('reset')+highlight_module(moduleid[-1])
    if short:
        return moduleid[-1]
    return ".".join(moduleid)

def register(mf):

    # precompute events
    events = {}
    for module_id in mf.modules_avail.keys():
        module_path = mf.modules_avail[module_id]["importpath"]

        # ignore self
        if __name__ == mf.getModuleId(module_id):
            continue

        # note all events
        module_events = mf.getModuleEvents(module_id)
        for e in module_events:
            if e in events:
                events[e].append(module_id)
            else:
                events[e] = [module_id]
                
    # register actual module
    mf.register_defaults({
        "event": "",
        "module": "",
        "long": False,
    })
    mf.register_helpers({
        "events": events
    })
    mf.register_event('init', init)


def init(state):

    # print it
    print()
    print(fg('yellow')+attr('underlined')+"Available events"+attr('reset'))
    event_list = state["events"].keys()
    if state["event"]:
        r = re.compile(".*"+state["event"]+".*")
        event_list = list(filter(lambda e: r.match(e[0]),event_list))
    if state["module"]:
        new_events = {}
        r = re.compile(".*"+state["module"]+".*")
        for e in state["events"].keys():
            modules = list(filter(r.match, state["events"][e]))
            if len(modules) > 0:
                new_events[e] = modules
        state["events"] = new_events
        event_list = list(filter(lambda e: e in state["events"],event_list))
    for e in sorted(event_list):
        unique_flag = "!" if e[1] else ">"
        print(fg('yellow')+unique_flag+' '+e[0]+attr('reset')+" used in "+", ".join([color_module(ev, short=not state["long"]) for ev in state["events"][e]])+attr('reset'))
    print()



from colored import fg, bg, attr

class state(dict):
    def __init__(self, module_name, state, state_default):
        self.all = state
        self.default = state_default
        self.module_id = module_name

    def scope(self, module_name, local=False):
        return state(self.module_id+"."+module_name if local else module_name, self.state, self.state_default)

    def __contains__(self, name):
        return self.module_id+"."+name in self.all if len(self.module_id) > 0 else name in self.all
    def __getitem__(self, name):
        return self.all[self.module_id+"."+name if len(self.module_id) > 0 else name]
    def __setitem__(self, name, val):
        self.all[self.module_id+"."+name if len(self.module_id) > 0 else name] = val

    def __getattribute__(self, name):
        return super().__getattribute__(name)

class like:
    def __init__(self, varname, alt, scope=None, scope_name=None):
        if scope_name is None:
            scope_name = scope
        global_varname = varname if scope is None else scope + "." + varname
        self.varname = scope_name+"."+varname if scope_name is not None else varname
        self.alt = alt
        self.fn = lambda state,event: state[global_varname] if global_varname in state else alt

    def __call__(self, state, event):
        return self.fn(state,event)

    def __str__(self):
        return attr('dim')+"'"+str(self.varname)+"' or '"+str(self.alt)+"' ⟶   "+attr('reset')+str(self.default)




from .dummy import dummy_fn
from .exceptions import RegisterError
import collections
import inspect
from functools import partial
from colored import fg, attr

class outervar():
    pass

class event_obj():
    def __init__(self, fn, unique, module, call_before_after=True):
        self.unique = unique
        self.call_before_after = call_before_after
        if unique:
            self.fn = fn
            self.modules = module
        else:
            self.fn = [fn]
            self.modules = [module]

class event(dict):
    def __init__(self, mf, optional=False):
        self._mf = mf
        self.optional_value = optional
        self.locals = {}

    def __getattr__(self, name):

        if name not in self._mf.event_objs:
            if not self.optional_value:
                raise AttributeError("The Event '%s' has not been registered yet." % name)
            fn_wrap = dummy_fn

        else:

            eobj = self._mf.event_objs[name]
            call_before_after = eobj.call_before_after

            # fn_wrap_scope creates a function wrap of fn that passes also state and event of eobj
            # additionally, if outervar is defined as a default, it queries that from the last outer scope
            def fn_wrap_scope(fn, state, event, module, needed_locals=None, miniflask_args=None, skip_twice=False, call_before_after=call_before_after):
                if needed_locals is None:
                    needed_locals = []
                if miniflask_args is None:
                    miniflask_args = []

                # get kwargs of fn wit outervar as default
                signature = inspect.signature(fn)
                needed_locals = [
                    k for k, v in signature.parameters.items()
                    if v.default is not inspect.Parameter.empty and v.default is outervar
                ]
                arg_names = [k for k, v in signature.parameters.items()]
                has_altfn = "altfn" in arg_names

                # automatic before/after events
                name_before = 'before_'+name
                name_after  = 'after_'+name
                has_before  = name_before in self._mf.event_objs and call_before_after
                has_after   = name_after in self._mf.event_objs and call_before_after
                fn_after    = getattr(self._mf.event, name_after) if has_after else None
                fn_before   = getattr(self._mf.event, name_before) if has_before else None

                # get index of "state" / "event"
                if len(arg_names) > 0:
                    if arg_names[0] == "state":
                        miniflask_args.append(state)
                        if len(arg_names) > 1 and arg_names[1] == "event":
                            miniflask_args.append(event)
                    elif arg_names[0] == "event":
                        miniflask_args.append(event)
                        if len(arg_names) > 1 and arg_names[1] == "state":
                            miniflask_args.append(state)

                # if no outervar found, just pass state and event
                if len(needed_locals) > 0:
                    def fn_wrap(*args, altfn=None, **kwargs):
                        outer_locals = {}
                        if skip_twice:
                            all_outer_locals = inspect.currentframe().f_back.f_back.f_locals
                        else:
                            all_outer_locals = inspect.currentframe().f_back.f_locals
                        outer_locals = {k: all_outer_locals[k] for k in needed_locals}
                        if has_altfn:
                            kwargs["altfn"] = altfn
                        if has_before:
                            for fn_b in fn_before.fns:
                                args, kwargs = fn_b(*args, **kwargs)
                        res = fn(*miniflask_args,*args,**outer_locals,**kwargs)
                        if has_after:
                            for fn_a in fn_after.fns:
                                res, args, kwargs = fn_a(res, *args, **kwargs)
                        return res
                elif len(miniflask_args) > 0:
                    def fn_wrap(*args, altfn=None, **kwargs):
                        if has_altfn:
                            kwargs["altfn"] = altfn
                        if has_before:
                            for fn_b in fn_before.fns:
                                args, kwargs = fn_b(*args, **kwargs)
                        res = fn(*miniflask_args, *args, **kwargs)
                        if has_after:
                            for fn_a in fn_after.fns:
                                res, args, kwargs = fn_a(res, *args, **kwargs)
                        return res
                else:
                    # it would be nice to let the user know, if the definition may be wrong at this point,
                    # however, we cannot know, if the call will contain the altfn-argument
                    # (up until we have a clean concept for this, we do not catch these definition errors)
                    # if not has_altfn and self.optional_value:
                    #     raise RegisterError(("The event %s was called using `event.optional(..., alftn=...)`, but the function does not catch this argument.\n\n"+fg('red')+"Possible Solutions:"+attr('reset')+"\n  - add `**kwargs` or `altfn=None` to your event-function definition.\n  - Alternatively, add either `event` or `state` or both to the event-function definition. In that case miniflask can catch altfn itself, however, this may adversely affect performance if this function is callled often.") % (fg('red')+name+attr('reset')))
                    return fn
                return fn_wrap

            if eobj.unique:
                fn_wrap = fn_wrap_scope(eobj.fn, eobj.modules.state, eobj.modules.event, eobj.modules)
                setattr(fn_wrap, 'modules', [eobj.modules.module_id])
                setattr(fn_wrap, 'fns', [fn_wrap])
            else:
                def multiple_fn_wrap_scope(orig_fns, modules=eobj.modules):
                    fns = [fn_wrap_scope(fn, state=module.state, event=module.event, module=module, skip_twice=True) for fn, module in zip(orig_fns,modules)]
                    def fn_wrap(*args, altfn=None, **kwargs):
                        results = []
                        for i,fn in enumerate(fns):
                            results.append(fn(*args, **kwargs))
                        return results
                    return fn_wrap, fns
                fn_wrap, fns = multiple_fn_wrap_scope(eobj.fn)
                setattr(fn_wrap, 'modules', [m.module_id for m in eobj.modules])
                setattr(fn_wrap, 'fns', fns)

        setattr(self, name, fn_wrap)
        return fn_wrap


